_allowed_ascii_chars = (
    'abcdefghijklmnopqrstuvwxyz'
    'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    '0123456789'
    '-_'
)


def is_ascii_str(s):
    if not isinstance(s, str):
        return False
    for c in s:
        if c not in _allowed_ascii_chars:
            return False
    return True



class AuthTokenError(Exception):
    """
        Base exception used for all auth token errors.
    """
    pass


class ConfigError(AuthTokenError):
    """
        Raised when an error occurs with a configuration parameter.
    """
    pass


class InvalidValueError(AuthTokenError):
    """
        Raised when metadata used to create or verify a token is invalid.
    """
    pass


class InvalidTokenError(AuthTokenError):
    """
        Parent class for all errors that can occur when validating a token.
    """
    pass


class TokenExpiredError(InvalidTokenError):
    """
        Raised when a token has expired.
    """
    pass


class TokenSignatureError(InvalidTokenError):
    """
        Raised when a token has an invalid signature.
    """
    pass



import os
import string
import unittest
from time import sleep
from hashlib import sha256, sha512
import handshake
from handshake import utils as handshake_utils


class HandshakeTestCase(unittest.TestCase):

    maxDiff = None

    def test_utils_is_ascii_str(self):
        # Invalid types
        self.assertFalse(handshake_utils.is_ascii_str(b''))
        self.assertFalse(handshake_utils.is_ascii_str(0))
        self.assertFalse(handshake_utils.is_ascii_str({}))
        self.assertFalse(handshake_utils.is_ascii_str(object()))
        # Invalid example chars
        self.assertFalse(handshake_utils.is_ascii_str(' '))
        self.assertFalse(handshake_utils.is_ascii_str('!'))
        self.assertFalse(handshake_utils.is_ascii_str('@'))
        self.assertFalse(handshake_utils.is_ascii_str('*'))
        self.assertFalse(handshake_utils.is_ascii_str(':'))
        # Valid chars
        for c in string.ascii_lowercase:
            self.assertTrue(handshake_utils.is_ascii_str(c))
        for c in string.ascii_uppercase:
            self.assertTrue(handshake_utils.is_ascii_str(c))
        for c in string.digits:
            self.assertTrue(handshake_utils.is_ascii_str(c))
        for c in '_-':
            self.assertTrue(handshake_utils.is_ascii_str(c))

    def test_interface(self):
        # No args
        with self.assertRaises(handshake.errors.ConfigError):
            handshake.AuthToken()
        # Too small a secret
        with self.assertRaises(handshake.errors.ConfigError):
            handshake.AuthToken('small')
        # Too large a secret
        with self.assertRaises(handshake.errors.ConfigError):
            handshake.AuthToken('massive' * 1024)
        # Invalid types
        with self.assertRaises(handshake.errors.ConfigError):
            handshake.AuthToken(0)
        with self.assertRaises(handshake.errors.ConfigError):
            handshake.AuthToken({})
        with self.assertRaises(handshake.errors.ConfigError):
            handshake.AuthToken(object())
        # Invalid callbacks
        with self.assertRaises(handshake.errors.ConfigError):
            handshake.AuthToken('secret secret secret', hashfunc=0)
        with self.assertRaises(handshake.errors.ConfigError):
            handshake.AuthToken('secret secret secret', hashfunc='')
        with self.assertRaises(handshake.errors.ConfigError):
            handshake.AuthToken('secret secret secret', hashfunc=b'')
        with self.assertRaises(handshake.errors.ConfigError):
            handshake.AuthToken('secret secret secret', hashfunc=lambda x: x)
        # Invalid random_len
        with self.assertRaises(handshake.errors.ConfigError):
            handshake.AuthToken('secret secret secret', random_len=0)
        with self.assertRaises(handshake.errors.ConfigError):
            handshake.AuthToken('secret secret secret', random_len=9999)
        # Valid initialisation
        handshake.AuthToken('secret secret secret')
        handshake.AuthToken(b'secret secret secret')
        handshake.AuthToken('secret secret secret', hashfunc=sha256)
        handshake.AuthToken('secret secret secret', random_len=128)

    def test_create_token(self):
        random_secret = os.urandom(256)
        auth_tokens = handshake.AuthToken(random_secret)
        # Invalid params
        with self.assertRaises(handshake.errors.InvalidValueError):
            auth_tokens.create(' ')
        with self.assertRaises(handshake.errors.InvalidValueError):
            auth_tokens.create('!')
        with self.assertRaises(handshake.errors.InvalidValueError):
            auth_tokens.create('@')
        with self.assertRaises(handshake.errors.InvalidValueError):
            auth_tokens.create('*')
        with self.assertRaises(handshake.errors.InvalidValueError):
            auth_tokens.create(':')
        # Too much metadata
        with self.assertRaises(handshake.errors.InvalidValueError):
            too_much_metadata = ['very long'] * 1024
            auth_tokens.create(*too_much_metadata)
        # Bare token
        test_token = auth_tokens.create()
        self.assertEqual(test_token[0], ':')
        token_parts = test_token.split(':')
        self.assertEqual(len(token_parts), 4)
        int(token_parts[1])                         # timestamp
        self.assertEqual(len(token_parts[2]), 64)   # random
        self.assertEqual(len(token_parts[3]), 64)   # signature
        # Arbitrary parameters
        test_token = auth_tokens.create('test', 'params')
        token_parts = test_token.split(':')
        self.assertEqual(len(token_parts), 5)
        self.assertEqual(token_parts[0], 'test')
        self.assertEqual(token_parts[1], 'params')
        int(token_parts[2])                         # timestamp
        self.assertEqual(len(token_parts[3]), 64)   # random
        self.assertEqual(len(token_parts[4]), 64)   # signature
        # Using a different hash function
        auth_tokens = handshake.AuthToken(random_secret, hashfunc=sha512)
        test_token = auth_tokens.create('test', 'params')
        token_parts = test_token.split(':')
        self.assertEqual(len(token_parts), 5)
        self.assertEqual(token_parts[0], 'test')
        self.assertEqual(token_parts[1], 'params')
        int(token_parts[2])                         # timestamp
        self.assertEqual(len(token_parts[3]), 128)  # random
        self.assertEqual(len(token_parts[4]), 128)  # signature

    def test_verify_token(self):
        random_secret = os.urandom(256)
        auth_tokens = handshake.AuthToken(random_secret)
        # Invalid token, modify the signature slightly
        test_token = auth_tokens.create()
        test_token_list = list(test_token)
        test_token_list[-1] = '1' if test_token_list[-1] == '0' else '0'
        test_token = ''.join(test_token_list)
        with self.assertRaises(handshake.errors.TokenSignatureError):
            auth_tokens.verify(test_token)
        # Valid token
        test_token = auth_tokens.create()
        verify_result = auth_tokens.verify(test_token)
        self.assertEqual(verify_result, ())
        # Arbitrary parameters
        test_token = auth_tokens.create('test', 'params')
        verify_result = auth_tokens.verify(test_token)
        self.assertEqual(verify_result, ('test', 'params'))
        # Valid time range
        test_token = auth_tokens.create('time', 'range')
        verify_result = auth_tokens.verify(test_token, time_range=100)
        self.assertEqual(verify_result, ('time', 'range'))
        # Invalid time range
        test_token = auth_tokens.create('time', 'range')
        sleep(2)  # Sleep for more than the specified time_range
        with self.assertRaises(handshake.errors.TokenExpiredError):
            verify_result = auth_tokens.verify(test_token, time_range=1)
        # Fixed, known token for full test
        auth_tokens = handshake.AuthToken('secret secret secret')
        test_token = ('test:'
                      '1:'
                      '00000000000000000000000000000000'
                      '00000000000000000000000000000000:'
                      'e7fb75d1239b69a9fbe7912719d70602'
                      '020322156d1bc8e3c31f3b2870180279')
        verify_result = auth_tokens.verify(test_token)
        self.assertEqual(verify_result, ('test',))



import os
import hmac
from time import time
from hashlib import sha256
from . import errors, utils


class AuthToken:

    MIN_SECRET_LEN = 16
    MAX_SECRET_LEN = 1024
    MAX_TOKEN_LEN = 2048

    def __init__(self, secret='', hashfunc=sha256, random_len=256):
        if isinstance(secret, str):
            secret = secret.encode()
        if not isinstance(secret, bytes):
            raise errors.ConfigError(f'secret must be a str or '
                              f'bytes, got: {type(secret)}')
        secret_len = len(secret)
        if (secret_len < self.MIN_SECRET_LEN or
            secret_len > self.MAX_SECRET_LEN):
            raise errors.ConfigError(f'secret must be between '
                                     f'{self.MIN_SECRET_LEN} and '
                                     f'{self.MAX_SECRET_LEN} chars or bytes '
                                     f'in length, got: {secret_len}')
        self.secret = secret
        if not callable(hashfunc):
            raise errors.ConfigError(f'hashfunc must be a callable, got: '
                              f'{type(hashfunc)}')
        try:
            h = hashfunc(b'').hexdigest()
        except Exception as e:
            raise errors.ConfigError(f'hash function is invalid: {hashfunc}, '
                                     f'the callable must return an object '
                                     f'with a hexdigest() method')
        self.hashfunc = hashfunc
        if not isinstance(random_len, int):
            raise errors.ConfigError(f'random_len must be an int, '
                              f'got: {type(random_len)}')
        if (random_len < self.MIN_SECRET_LEN or
            random_len > self.MAX_SECRET_LEN):
            raise errors.ConfigError(f'random_len must be between '
                                     f'{self.MIN_SECRET_LEN} and '
                                     f'{self.MAX_SECRET_LEN}, '
                                     f'got: {random_len}')
        self.random_len = random_len

    def create(self, *metadata):
        params = []
        for i, param in enumerate(metadata):
            if not utils.is_ascii_str(param):
                raise errors.InvalidValueError(f'metadata[{i}] contains '
                                               f'invalid characters, only '
                                               f'a-z A-Z 0-9 underscores '
                                               f'and hyphens are allowed, '
                                               f'got: {param}')
            params.append(param)
        param_str = ':'.join(params)
        ts = str(int(time()))
        random_str = self.hashfunc(os.urandom(self.random_len)).hexdigest()
        msg = f'{param_str}:{ts}:{random_str}'
        if len(msg) > self.MAX_TOKEN_LEN:
            raise errors.InvalidValueError(f'token must be less than '
                                           f'{self.MAX_TOKEN_LEN} characters '
                                           f'when generated (try specifying '
                                           f'shorter metadata), token was '
                                           f'{len(msg)} characters')
        h = hmac.new(self.secret, msg=msg.encode(), digestmod=self.hashfunc)
        return f'{msg}:{h.hexdigest()}'

    def verify(self, token, time_range=0):
        if not isinstance(token, str):
            raise errors.InvalidTokenError(f'token must be a str, '
                                           f'got: {type(token)}')
        if len(token) > self.MAX_TOKEN_LEN:
            raise errors.InvalidTokenError(f'token must be '
                                           f'{self.MAX_TOKEN_LEN} chars or '
                                           f'less, got: {len(token)}')
        parts = token.split(':')
        if len(parts) < 4:
            raise errors.InvalidTokenError(f'tokens must contain at least 4 '
                                           f'parts separated by colons (:), '
                                           f'got: {token}')
        if not isinstance(time_range, int):
            raise errors.InvalidValueError(f'time_range must be an int, '
                                           f'got: {type(time_range)}')
        if time_range < 0:
            raise errors.InvalidValueError(f'time_range must be postive, '
                                           f'got: {time_range}')
        timestamp, random, signature = parts[-3], parts[-2], parts[-1]
        try:
            timestamp = int(timestamp)
        except (ValueError, TypeError) as e:
            raise errors.InvalidTokenError(f'token has an invalid timestamp, '
                                           f'unable to parse {timestamp} as '
                                           f'an int: {e}')
        metadata = parts[0:-3]
        if time_range > 0:
            local_ts = int(time())
            min_drift = local_ts - time_range
            max_drift = local_ts + time_range
            if timestamp < min_drift or timestamp > max_drift:
                raise errors.TokenExpiredError(f'token has expired, timestamp '
                                               f'{timestamp} with time_range '
                                               f'of {time_range} and local '
                                               f'time of {local_ts}')
        metadata_str = ':'.join(metadata)
        msg = f'{metadata_str}:{timestamp}:{random}'
        h = hmac.new(self.secret, msg=msg.encode(), digestmod=self.hashfunc)
        test_signature = h.hexdigest()
        if not hmac.compare_digest(test_signature, signature):
            raise errors.TokenSignatureError(f'token has an invalid signature')
        return () if metadata == [''] else tuple(metadata)
